from django.urls import path, include
from django.views.generic.base import TemplateView
from .views import (
    UserKeyListView,
    UserKeyDeleteView,
    MFAAdminLoginView,
    MFAAdminSecondFactorView,
)


urlpatterns = [
    path("login/", MFAAdminLoginView.as_view()),
    path("login/mfa/", MFAAdminSecondFactorView.as_view(), name="login-mfa"),
    path("multi_factor_auth/", UserKeyListView.as_view(), name="user-key-list"),
    path(
        "multi_factor_auth/<pk>/delete/",
        UserKeyDeleteView.as_view(),
        name="user-key-delete",
    ),
    path(
        "multi_factor_auth/fido2/",
        TemplateView.as_view(template_name="django_rest_mfa/fido2_add.html"),
        name="fido2-add",
    ),
    path(
        "multi_factor_auth/otp/",
        TemplateView.as_view(template_name="django_rest_mfa/otp_add.html"),
        name="otp-add",
    ),
]



from django.conf import settings
from fido2.client import ClientData
from fido2.ctap2 import AttestedCredentialData, AttestationObject, AuthenticatorData
from fido2.server import Fido2Server, PublicKeyCredentialRpEntity
from fido2.utils import websafe_decode, websafe_encode


def get_server():
    rp = PublicKeyCredentialRpEntity(settings.FIDO_SERVER_ID, settings.MFA_SERVER_NAME)
    return Fido2Server(rp)


def get_user_credentials(user):
    credentials = []
    for uk in user.userkey_set.filter(key_type="FIDO2"):
        credentials.append(
            AttestedCredentialData(websafe_decode(uk.properties.get("device")))
        )
    return credentials


def generate_fido2_registration(user):
    """Returns registration_data, state"""
    server = get_server()
    return server.register_begin(
        {
            "id": str(user.pk).encode(),
            "name": user.get_full_name(),
            "displayName": user.get_username(),
        },
        get_user_credentials(user),
    )


def fido2_complete_registration(client_data_json, attestation_object, fido_state):
    """Returns properties"""
    client_data = ClientData(client_data_json)
    att_obj = AttestationObject(attestation_object)
    server = get_server()
    auth_data = server.register_complete(
        fido_state,
        client_data,
        att_obj,
    )
    encoded = websafe_encode(auth_data.credential_data)
    return {
        "device": encoded,
        "type": att_obj.fmt,
    }


def fido2_begin_authenticate(user):
    """Returns auth_data, fido_state"""
    server = get_server()
    credentials = get_user_credentials(user)
    return server.authenticate_begin(credentials)


def fido2_complete_authenticate(
    user, credential_id, client_data_json, authenticator_data, signature, fido_state
):
    server = get_server()
    credentials = get_user_credentials(user)

    client_data = ClientData(client_data_json)
    auth_data = AuthenticatorData(authenticator_data)

    cred = server.authenticate_complete(
        fido_state,
        credentials,
        credential_id,
        client_data,
        auth_data,
        signature,
    )

    keys = user.userkey_set.filter(key_type="FIDO2")
    for key in keys:
        if (
            AttestedCredentialData(
                websafe_decode(key.properties["device"])
            ).credential_id
            == cred.credential_id
        ):
            return key


# Generated by Django 3.2.3 on 2021-05-25 01:12

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserKey',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=255)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('key_type', models.CharField(choices=[('FIDO2', 'Fido2'), ('TOTP', 'Totp')], max_length=25)),
                ('properties', models.JSONField(default=dict)),
                ('last_used', models.DateTimeField(blank=True, default=None, null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from django.contrib.auth import get_user_model
from django.urls import reverse
from model_bakery import baker
from rest_framework.test import APITestCase
import pyotp
from fido2 import cbor

from .totp import generate_totp_key
from .models import UserKey
from .helpers import set_expirable_var
from .constants import USER_PK_SESSION_KEY

from django_rest_mfa.fido2 import fido2_complete_registration


class UserKeyTests(APITestCase):
    def setUp(self):
        self.user = baker.make(get_user_model())
        self.client.force_login(self.user)

    def test_list(self):
        url = reverse("user-keys-list")
        key = baker.make("django_rest_mfa.UserKey", user=self.user, name="one")
        key2 = baker.make("django_rest_mfa.UserKey", name="two")
        res = self.client.get(url)
        self.assertContains(res, key.name)
        self.assertNotContains(res, key2.name)

    def test_delete(self):
        key = baker.make("django_rest_mfa.UserKey", user=self.user)
        url = reverse("user-keys-detail", kwargs={"pk": key.pk})
        res = self.client.delete(url)
        self.assertEqual(res.status_code, 204)

    def test_totp(self):
        url = reverse("user-keys-list") + "totp/"
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)

        totp = pyotp.TOTP(res.data["secret_key"])
        data = {"secret_key": res.data["secret_key"], "answer": totp.now()}
        res = self.client.post(url, data, format="json")
        self.assertEqual(res.status_code, 200)
        self.assertEqual(UserKey.objects.count(), 1)

    def test_invalid_totp(self):
        url = reverse("user-keys-list") + "totp/"
        res = self.client.get(url)
        totp = pyotp.TOTP(res.data["secret_key"])
        data = {"secret_key": res.data["secret_key"], "answer": "111111"}
        res = self.client.post(url, data, format="json")
        self.assertEqual(res.status_code, 400)
        self.assertEqual(UserKey.objects.count(), 0)

    def test_fido2(self):
        url = reverse("user-keys-list") + "fido2/"
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)
        data = {
            "attestationObject": b"",
            "clientDataJSON": b"",
            "name": self.user.get_username(),
        }

        # Not sure how to test this
        # res = self.client.post(url, data, format="json")
        # self.assertEqual(res.status_code, 200)
        # self.assertEqual(UserKey.objects.count(), 1)

    def test_invalid_fido2(self):
        url = reverse("user-keys-list") + "fido2/"
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)

        data = {
            "attestationObject": b"cf",
            "clientDataJSON": b'{"challenge":"P","clientExtensions":{},"hashAlgorithm":"SHA-256","origin":"https://localhost:8000","type":"webauthn.create"}',
            "name": self.user.get_username(),
        }
        res = self.client.post(url, cbor.encode(data), format="json")
        self.assertContains(res, "Invalid data", status_code=400)
        self.assertEqual(UserKey.objects.count(), 0)

    def test_trusted_device(self):
        url = reverse("user-keys-list") + "trusted_device/"
        res = self.client.get(url, HTTP_USER_AGENT="Mozilla/5.0")
        self.assertEqual(res.status_code, 200)

        data = {"key": res.data["key"], "name": res.data["user_agent"]}
        res = self.client.post(url, data, HTTP_USER_AGENT="Mozilla/5.0")
        self.assertEqual(res.status_code, 200)
        self.assertEqual(UserKey.objects.count(), 1)

    def test_backup_codes(self):
        url = reverse("user-keys-list") + "backup_codes/"
        data = {"name": "codes"}
        res = self.client.post(url, data)
        self.assertEqual(res.status_code, 200)
        self.assertEqual(UserKey.objects.count(), 1)

        # Should always replace codes
        res = self.client.post(url, data)
        self.assertEqual(res.status_code, 200)
        self.assertEqual(UserKey.objects.count(), 1)


class AuthenticateTests(APITestCase):
    def setUp(self):
        self.user = baker.make(get_user_model())

    def test_authenticate_totp(self):
        url = reverse("authenticate-totp")
        totp_key = generate_totp_key(self.user)
        totp = pyotp.TOTP(totp_key["secret_key"])
        key = baker.make(
            "django_rest_mfa.UserKey",
            user=self.user,
            key_type=UserKey.KeyType.TOTP,
            properties={"secret_key": totp_key["secret_key"]},
        )
        session = self.client.session
        set_expirable_var(session, USER_PK_SESSION_KEY, self.user.pk)
        session.save()
        data = {"otp": totp.now()}
        res = self.client.post(url, data)
        self.assertEqual(res.status_code, 200)

        self.client.logout()
        session = self.client.session
        set_expirable_var(session, USER_PK_SESSION_KEY, self.user.pk)
        session.save()
        res = self.client.post(url, data)
        self.assertContains(res, "OTP already used", status_code=400)

        data = {"otp": "111111"}
        res = self.client.post(url, data)
        self.assertContains(res, "Invalid TOTP code", status_code=400)

        # Refresh session
        self.client.login()
        self.client.logout()
        data = {"otp": totp.now()}
        res = self.client.post(url, data)
        self.assertContains(res, "Unknown user", status_code=400)

    def test_authenticate_fido2(self):
        url = reverse("authenticate-fido2")
        data = {}
        res = self.client.post(url, data)
        self.assertEqual(res.status_code, 400)

    def test_authenticate_backup_codes(self):
        url = reverse("authenticate-backup-codes")
        codes = [
            "H4DVCQLXLSRJ098Z",
            "8QRWNBO5WHFUJQC3",
            "GL2LQAQVVOQPNELK",
            "PM2CJ5J7X8BKAX3R",
            "5457MUT07EJWH56Y",
            "WFNBXBVKQ6V0J1ZT",
            "Z0QZNQHOUN6KSWOV",
            "Q5SO1TB8LKTHDWZN",
            "4S1F9QGOJ0HMYJFT",
            "IIJ4S9SO3KKZOX4L",
        ]
        key = baker.make(
            "django_rest_mfa.UserKey",
            user=self.user,
            key_type=UserKey.KeyType.BACKUP_CODES,
            properties={"codes": codes},
        )

        data = {"code": codes[0]}
        res = self.client.post(url, data)
        self.assertEqual(res.status_code, 400)

        session = self.client.session
        set_expirable_var(session, USER_PK_SESSION_KEY, self.user.pk)
        session.save()
        res = self.client.post(url, data)
        self.assertEqual(res.status_code, 200)
        key.refresh_from_db()
        self.assertNotIn(codes[0], key.properties.get("codes"))

        data = {"code": codes[0]}
        res = self.client.post(url, data)
        self.assertEqual(res.status_code, 400)

        data = {"code": codes[1]}
        res = self.client.post(url, data)
        self.assertEqual(res.status_code, 200)


from django.contrib.auth import get_user_model, login
from rest_framework import viewsets, mixins, permissions, renderers, exceptions, parsers
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.decorators import action
from .constants import USER_PK_SESSION_KEY, USER_AUTH_BACKEND
from .models import UserKey
from .serializers import (
    UserKeySerializer,
    OTPUserKeySerializer,
    VerifyOTPSerializer,
    FIDO2RegistrationSerializer,
    FIDO2AuthenticateSerializer,
    TrustedDeviceSerializer,
    BackupCodesSerializer,
    BackupCodeAuthSerializer
)
from .renderers import CBORRenderer, Base64JSONRenderer
from .parsers import CBORParser, Base64JSONParser
from .helpers import set_expirable_var, get_expirable_var
from .totp import generate_totp_key, verify_totp
from .fido2 import (
    generate_fido2_registration,
    fido2_begin_authenticate,
    fido2_complete_authenticate,
)
from .trusted_device import generate_random_key, generate_ua_short


class UserKeyViewSet(
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    mixins.DestroyModelMixin,
    viewsets.GenericViewSet,
):
    """Manage User Keys, including view and delete."""

    queryset = UserKey.objects.none()
    serializer_class = UserKeySerializer
    permission_classes = [permissions.IsAuthenticated]
    renderer_classes = [
        Base64JSONRenderer,
        renderers.BrowsableAPIRenderer,
        CBORRenderer,
        renderers.MultiPartRenderer,
    ]
    parser_classes = [Base64JSONParser, CBORParser, parsers.MultiPartParser]

    def get_queryset(self):
        return self.request.user.userkey_set.all()

    def get_serializer_class(self):
        if self.action == "totp":
            return OTPUserKeySerializer
        if self.action == "fido2":
            return FIDO2RegistrationSerializer
        if self.action == "trusted_device":
            return TrustedDeviceSerializer
        if self.action == "backup_codes":
            return BackupCodesSerializer
        return super().get_serializer_class()

    @action(detail=False, methods=["get", "post"], name="Set up TOTP")
    def totp(self, request):
        """Set up TOTP (Time-Based One Time Password)"""
        if request.method == "GET":
            result = generate_totp_key(request.user)
            return Response(result)

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save(
            key_type=UserKey.KeyType.TOTP, user=self.request.user, name="TOTP"
        )
        return Response(serializer.data)

    @action(detail=False, methods=["get", "post"], name="Add new FIDO2 Key")
    def fido2(self, request):
        if request.method == "GET":
            registration_data, state = generate_fido2_registration(request.user)
            set_expirable_var(request.session, "fido_state", state)
            return Response(registration_data)

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save(key_type=UserKey.KeyType.FIDO2, user=self.request.user)
        return Response(serializer.data)
    
    @action(detail=False, methods=["get", "post"], name="Add a trusted device")
    def trusted_device(self, request):
        if request.method == "GET":
            key = generate_random_key()
            user_agent = generate_ua_short(request.META["HTTP_USER_AGENT"])
            return Response({
                "key": key,
                "user_agent": user_agent
            })

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save(key_type=UserKey.KeyType.DEVICE, user=self.request.user)
        return Response(serializer.data)

    @action(detail=False, methods=["post"], name="Set backup codes")
    def backup_codes(self, request):
        """
        Make one time backup codes for user
        Only one set up codes may exist, requesting new ones will delete existing
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save(key_type=UserKey.KeyType.DEVICE, user=self.request.user)
        return Response(serializer.data)


class MFAAuthenticateAPIView(APIView):
    permission_classes = [permissions.AllowAny]

    def get_user(self, request):
        user = request.user
        if not user.is_authenticated:
            User = get_user_model()
            try:
                user_pk = get_expirable_var(request.session, USER_PK_SESSION_KEY)
                auth_backend = get_expirable_var(request.session, USER_AUTH_BACKEND)
                user = User.objects.get(pk=user_pk)
                if auth_backend:
                    user.backend=auth_backend
            except User.DoesNotExist:
                raise exceptions.ValidationError("Unknown user")
        return user
    
    def success(self, request, user, key):
        key.login_success(request.session)
        if not request.user.is_authenticated:
            res = login(request, user)



class FIDO2Authenticate(MFAAuthenticateAPIView):
    renderer_classes = [
        Base64JSONRenderer,
        renderers.BrowsableAPIRenderer,
        CBORRenderer,
    ]
    parser_classes = [Base64JSONParser, CBORParser]

    def get(self, request):
        user = self.get_user(request)
        auth_data, state = fido2_begin_authenticate(user)
        set_expirable_var(request.session, "fido_state", state)
        return Response(auth_data)

    def post(self, request):
        user = self.get_user(request)
        serializer = FIDO2AuthenticateSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = serializer.validated_data

        try:
            key = fido2_complete_authenticate(
                user,
                data["credentialId"],
                data["clientDataJSON"],
                data["authenticatorData"],
                data["signature"],
                get_expirable_var(request.session, "fido_state"),
            )
        except ValueError as exc:
            raise exceptions.ValidationError(str(exc))

        self.success(request, user, key)
        return Response()


class TOTPAuthenticate(MFAAuthenticateAPIView):
    def post(self, request):
        user = self.get_user(request)
        serializer = VerifyOTPSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        otp = serializer.validated_data["otp"]
        try:
            key = user.userkey_set.get(key_type=UserKey.KeyType.TOTP)
        except UserKey.DoesNotExist:
            raise exceptions.ValidationError("TOTP is not enabled for user")

        if key.properties.get("last_otp") == otp:
            raise exceptions.ValidationError("OTP already used")

        success = verify_totp(
            key.properties.get("secret_key"), otp, 1
        )
        if not success:
            raise exceptions.ValidationError("Invalid TOTP code")
        key.properties["last_otp"] = otp
        self.success(request, user, key)
        return Response()


class BackupCodesAuthenticate(MFAAuthenticateAPIView):
    def post(self, request):
        user = self.get_user(request)
        serializer = BackupCodeAuthSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        code = serializer.validated_data.get("code")
        key = user.userkey_set.filter(key_type=UserKey.KeyType.BACKUP_CODES).first()
        if not key or code not in key.properties.get("codes"):
            raise exceptions.ValidationError("Backup code not found")
        self.success(request, user, key)

        # Remove used code
        key.properties["codes"].remove(code)
        key.save()
        return Response()



from django.conf import settings
import pyotp
from .models import UserKey


def generate_totp_key(user):
    secret_key = pyotp.random_base32()
    totp = pyotp.TOTP(secret_key)
    uri = totp.provisioning_uri(
        user.get_username(), issuer_name=settings.MFA_SERVER_NAME
    )
    return {"provisioning_uri": uri, "secret_key": secret_key}


def verify_totp(secret_key: str, answer: str, valid_window=0):
    totp = pyotp.TOTP(secret_key)
    return totp.verify(answer, valid_window=valid_window)



import string
import secrets
from user_agents import parse


def generate_random_key():
    return "".join(
        secrets.choice(string.ascii_uppercase + string.digits) for _ in range(20)
    )


def generate_ua_short(ua_string: str) -> str:
    """Generate a short user agent of Device / Operating System / Browser"""
    user_agent = parse(ua_string)
    return "{device} / {os} / {browser}".format(
        device=user_agent.get_device(),
        os=user_agent.get_os(),
        browser=user_agent.browser.family,
    )



from rest_framework.response import Response
from dj_rest_auth.views import LoginView
from ..helpers import has_mfa


class MFALoginView(LoginView):
    """
    MFA Login View for session based MFA aware authentication
    If the user has MFA enabled, mark the session but don't log in
    Otherwise, use default behavior

    On a successful login with MFA enabled, return {"requires_mfa": true} with status code 200
    """

    def process_login(self):
        if has_mfa(self.request, self.user):
            self.user.mfa = True
        else:
            self.user.mfa = False
            super().process_login()

    def get_response(self):
        if self.user.mfa:
            return Response({"requires_mfa": True})
        return super().get_response()



from rest_framework import serializers, exceptions
from .totp import verify_totp
from .fido2 import fido2_complete_registration
from .trusted_device import generate_ua_short
from .models import UserKey
from .fields import BinaryField
from .helpers import get_expirable_var


class BaseUserKeySerializer(serializers.ModelSerializer):
    class Meta:
        model = UserKey
        extra_kwargs = {"properties": {"write_only": True}}


class UserKeySerializer(BaseUserKeySerializer):
    class Meta(BaseUserKeySerializer.Meta):
        fields = (
            "id",
            "user",
            "name",
            "created",
            "key_type",
            "last_used",
            "properties",
        )


class VerifyOTPSerializer(serializers.Serializer):
    otp = serializers.CharField()


class BackupCodeAuthSerializer(serializers.Serializer):
    code = serializers.CharField(max_length=16, min_length=16)


class OTPUserKeySerializer(BaseUserKeySerializer):
    answer = serializers.CharField(write_only=True)
    secret_key = serializers.RegexField(
        r"^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$",
        write_only=True,
    )

    class Meta(BaseUserKeySerializer.Meta):
        fields = ("id", "name", "answer", "secret_key")
        read_only_fields = ("name",)

    def validate(self, data):
        user = self.context["request"].user
        if user.userkey_set.filter(key_type=UserKey.KeyType.TOTP).exists():
            raise serializers.ValidationError("User may only set up TOTP once")
        answer = data.pop("answer")

        if not verify_totp(data["secret_key"], answer, valid_window=60):
            raise serializers.ValidationError("Invalid Verification Answer")

        return data

    def create(self, validated_data):
        validated_data["properties"] = {"secret_key": validated_data.pop("secret_key")}
        return super().create(validated_data)


class FIDO2RegistrationSerializer(BaseUserKeySerializer):
    """
    attestationObject and clientDataJSON are specified to be ArrayBuffer and not strings.
    https://developers.yubico.com/WebAuthn/WebAuthn_Developer_Guide/WebAuthn_Client_Registration.html
    """

    attestationObject = BinaryField(write_only=True)
    clientDataJSON = BinaryField(write_only=True)

    class Meta(BaseUserKeySerializer.Meta):
        fields = ("id", "name", "attestationObject", "clientDataJSON")

    def create(self, validated_data):
        try:
            properties = fido2_complete_registration(
                validated_data.pop("clientDataJSON"),
                validated_data.pop("attestationObject"),
                get_expirable_var(self.context["request"].session, "fido_state"),
            )
        except ValueError as exc:
            raise exceptions.ValidationError(str(exc))
        except TypeError:
            raise exceptions.ValidationError("Invalid data")

        validated_data["properties"] = properties
        return super().create(validated_data)


class FIDO2AuthenticateSerializer(serializers.Serializer):
    credentialId = BinaryField()
    authenticatorData = BinaryField()
    clientDataJSON = BinaryField()
    signature = BinaryField()


class TrustedDeviceSerializer(BaseUserKeySerializer):
    key = serializers.CharField(write_only=True, min_length=20, max_length=20)

    class Meta(BaseUserKeySerializer.Meta):
        fields = ("id", "name", "key")

    def create(self, validated_data):
        validated_data["properties"] = {
            "key": validated_data.pop("key"),
            "ua": generate_ua_short(self.context["request"].META["HTTP_USER_AGENT"]),
        }
        return super().create(validated_data)


class BackupCodesSerializer(BaseUserKeySerializer):
    codes = serializers.SerializerMethodField()

    class Meta(BaseUserKeySerializer.Meta):
        fields = ("id", "name", "codes")

    def create(self, validated_data):
        return UserKey.make_new_backup_codes_for_user(
            validated_data.get("user"), validated_data.get("name")
        )

    def get_codes(self, obj):
        return obj.properties.get("codes")



from rest_framework import renderers
from rest_framework.utils.encoders import JSONEncoder
from fido2.utils import websafe_encode, websafe_decode
from fido2 import cbor


class Base64JSONEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            return websafe_encode(obj)
        return super().default(obj)


class CBORRenderer(renderers.BaseRenderer):
    """
    Concise Binary Object Representation Renderer
    """

    format = 'cbor'
    media_type = "application/octet-stream"

    def render(self, data, media_type=None, renderer_context=None):
        return cbor.encode(data)


class Base64JSONRenderer(renderers.JSONRenderer):
    encoder_class = Base64JSONEncoder



from django.contrib.auth.views import LoginView
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect
from django.utils.decorators import method_decorator
from django.urls import reverse_lazy, reverse
from django.views.generic import DeleteView, TemplateView
from django.views.generic.list import ListView
from django_rest_mfa.models import UserKey
from django_rest_mfa.helpers import has_mfa


@method_decorator(login_required, name='dispatch')
class UserKeyListView(ListView):
    model = UserKey

    def get_queryset(self):
        return super().get_queryset().filter(user=self.request.user)
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["has_otp"] = self.get_queryset().filter(key_type=UserKey.KeyType.TOTP).exists()
        return context


@method_decorator(login_required, name='dispatch')
class UserKeyDeleteView(DeleteView):
    model = UserKey
    success_url = reverse_lazy("user-key-list")

    def get_queryset(self):
        return super().get_queryset().filter(user=self.request.user)



class MFAAdminLoginView(LoginView):
    template_name = "django_rest_mfa/login.html"

    def form_valid(self, form):
        if has_mfa(self.request, form.get_user()):
            user = form.get_user()
            next_url = ""
            next_param = self.request.GET.get("next")
            if next_param:
                next_url = "?next=" + next_param
            return HttpResponseRedirect(reverse("login-mfa") + next_url)
        return super().form_valid(form)


class MFAAdminSecondFactorView(TemplateView):
    template_name = "django_rest_mfa/second_factor.html"


from django.contrib.auth import authenticate, get_user_model
from rest_framework import serializers
from dj_rest_auth.serializers import LoginSerializer
from ..helpers import has_mfa


class NoopTokenSerializer(serializers.Serializer):
    """dj-rest-auth requires tokens, but we don't use them."""



from rest_framework import parsers
from .renderers import Base64JSONRenderer
from fido2 import cbor


class CBORParser(parsers.BaseParser):
    """
    Concise Binary Object Representation Parser
    """

    media_type = "application/cbor"

    def parse(self, stream, media_type=None, parser_context=None):
        content = stream.read()
        return cbor.decode(content)


class Base64JSONParser(parsers.JSONParser):
    renderer_class = Base64JSONRenderer



from django.urls import path, include
from rest_framework.routers import DefaultRouter
from .views import UserKeyViewSet, FIDO2Authenticate, TOTPAuthenticate, BackupCodesAuthenticate

router = DefaultRouter()
router.register(r"user_keys", UserKeyViewSet, basename="user-keys")

urlpatterns = [
    path("authenticate/fido2/", FIDO2Authenticate.as_view(), name="authenticate-fido2"),
    path("authenticate/totp/", TOTPAuthenticate.as_view(), name="authenticate-totp"),
    path("authenticate/backup_codes/", BackupCodesAuthenticate.as_view(), name="authenticate-backup-codes"),
    path("", include(router.urls)),
]



from django.conf import settings


def noop_token_creator(token_model, user, serializer):
    """Fake token creator to use sessions instead of tokens"""
    return None


class NoopModel:
    """
    dj-rest-auth doesn't officially support not having a Token model
    But we can use this nothing class instead
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_rest_mfa',
 'django_rest_mfa.mfa_admin',
 'django_rest_mfa.migrations',
 'django_rest_mfa.rest_auth_helpers']

package_data = \
{'': ['*'],
 'django_rest_mfa.mfa_admin': ['static/django_rest_mfa/*',
                               'templates/admin/*',
                               'templates/django_rest_mfa/*']}

install_requires = \
['dj-rest-auth>=2.1.7,<3.0.0',
 'djangorestframework>=3.12.4,<4.0.0',
 'fido2>=0.9.1,<0.10.0',
 'pyotp>=2.6.0,<3.0.0',
 'user-agents>=2.2.0,<3.0.0']

setup_kwargs = {
    'name': 'django-rest-mfa',
    'version': '0.1.2',
    'description': 'Django Rest Framework Endpoints for MFA including TOTP and FIDO2',
    'long_description': None,
    'author': 'David Burke',
    'author_email': 'david@burkesoftware.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# Generated by Django 3.2.4 on 2021-06-06 18:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_rest_mfa', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='userkey',
            name='key_type',
            field=models.CharField(choices=[('FIDO2', 'Fido2'), ('TOTP', 'Totp'), ('Trusted Device', 'Device'), ('Backup Codes', 'Backup Codes')], max_length=25),
        ),
    ]



import datetime
from .models import UserKey
from .constants import USER_PK_SESSION_KEY, USER_AUTH_BACKEND


def has_mfa(request, user):
    """Check if MFA is enabled for this user
    If yes, return all keys (which may be useful to redirect or show user options)
    In addition, mark session with user pk to mark successful initial factor login
    If no, return False (in which case, log in normally)
    """
    mfa_types = [UserKey.KeyType.FIDO2, UserKey.KeyType.TOTP]
    user_keys = user.userkey_set.filter(key_type__in=mfa_types)
    if user_keys.exists():
        # Mark session as successful step 1 authentication
        set_expirable_var(request.session, USER_PK_SESSION_KEY, user.pk)
        # Remember auth backend, needed to call django login function when multiple backends are used.
        set_expirable_var(request.session, USER_AUTH_BACKEND, user.backend)
        return user_keys
    return False


def set_expirable_var(session, var_name, value, expire_at=None):
    """Set session variable with expiration, expire_at defaults to 5 minutes."""
    if expire_at is None:
        expire_at = datetime.datetime.today() + datetime.timedelta(minutes=5)
    session[var_name] = {"value": value, "expire_at": expire_at.timestamp()}


def get_expirable_var(session, var_name, default=None):
    var = default
    if var_name in session:
        my_variable_dict = session.get(var_name, {})
        if my_variable_dict.get("expire_at", 0) > datetime.datetime.now().timestamp():
            return my_variable_dict.get("value")
        else:
            del session[var_name]
    return var



import secrets
import string
from django.conf import settings
from django.db import models
from django.utils import timezone


class UserKey(models.Model):
    class KeyType(models.TextChoices):
        FIDO2 = "FIDO2"
        TOTP = "TOTP"
        DEVICE = "Trusted Device"
        BACKUP_CODES = "Backup Codes"

    name = models.CharField(max_length=255, blank=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    created = models.DateTimeField(auto_now_add=True)
    key_type = models.CharField(max_length=25, choices=KeyType.choices)
    properties = models.JSONField(default=dict)
    last_used = models.DateTimeField(null=True, default=None, blank=True)

    def __str__(self):
        return "%s - %s" % (self.name, self.key_type)

    def set_session_mfa(self, session):
        mfa = {"verified": True, "method": self.key_type, "id": self.pk}
        session["mfa"] = mfa

    def login_success(self, session):
        self.last_used = timezone.now()
        self.save()
        self.set_session_mfa(session)

    @classmethod
    def make_new_backup_codes_for_user(cls, user, name="Backup Codes", amount=10, length=16):
        """Create new backup codes, replace any existing"""
        # Default 10 codes of 16 chars each
        backup_codes = [
            "".join(
                secrets.choice(string.ascii_uppercase + string.digits)
                for _ in range(length)
            )
            for _ in range(amount)
        ]
        user_key, created = cls.objects.get_or_create(
            user=user,
            key_type=cls.KeyType.BACKUP_CODES,
            defaults={"properties": {"codes": backup_codes}, "name": name},
        )
        if not created:
            user_key.properties = {"codes": backup_codes}
            user_key.name = name
            user_key.save()
        return user_key
